%%%%%%%%%%%%%%%%%%               WORKING VERSION                   %%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%                                 RELIE.RED                                    %%%
%%%                               Version 3.0                                    %%%
%%%                             Francesco Oliveri                                %%%
%%%             foliveri@unime.it; http://mat521.unime.it/oliveri                %%%
%%%                        Last Revision: August 9, 2021                         %%% 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% GLOBAL VARIABLES
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% INPUT VARIABLES
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% approxorder,arbelem,arborder,contact,diffeqs,freepars,generalequiv,,jetorder,
% lagrangian,leadders,nonclassical,nonpolyders,nonzeropars,qcond,uvar,variational,
% xvar,zorder, 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% OUTPUT VARIANLES
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% allinfinitesimals,allphi,arbconst,arbfun,fluxes,cogenerators,commtable,deteqs,
% distribution,generators,invcond,nzcomm,prolongation,symmetries 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% INTERMEDIATE VARIANLES
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% jet,jetapprox,jetequiv,jetsplit,solutionsdedv,steprelie
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
for all x let (sin(x))^2+(cos(x))^2 = 1 $
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% GLOBAL DEFAULT VARIABLES
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%                                   Initialization
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
jetorder:=0 $          % Jet order zero
xvar:={} $             % No independent variables
uvar:={} $             % No dependent variables
diffeqs:={} $          % No differential equations
leadders:={} $         % No leading derivatives
freepars:={} $         % No free parameters involved (for group classification)
nonzeropars:={} $      % No non-vanishing parameters
nonpolyders:={} $      % No derivatives occurring in non-polynomial form
arborder:=-1 $         % No equivalence transformations
arbelem:={} $          % No arbitrary elements
zorder:=0 $            % Maximum order of derivatives arbitrary elements depend on
generalequiv:=0 $      % the infinitesimals of xvar and uvar do not depend on arbelem
approxorder:=0 $       % Exact Lie symmetries; >=1 Order of approximate symmetries
nonclassical:=0 $      % Classical symmetries; >=1 for conditional symmetries
qcond:={} $            % List of indexes of Q-conditions (for conditional symmetries)
contact:=0 $           % Point symmetries
variational:=0 $       % No variational symmetries
lagrangian:={} $       % No Lagrangian defined
okinputrelie:={} $     % No inputs processed
steprelie:=0 $         % No processing done
isometries:=1 $        % Flag to generate Lie algebra of isometries
affine:=2 $            % Flag to generate Lie algebra of affine transformations
projective:=3 $        % Flag to generate Lie algebra of projective transformations
invcond:={} $          % No invariance conditions computed
deteqs:={} $           % No determining equations computed
symmetries:={} $       % No symmetries computed
generators:={} $       % No generators computed
cogenerators:={} $     % No cogenerators computed
fluxes:={} $           % No fluxes computed
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
load_package assist $
load_package crack $ % crack used for solving determining equations 
                     % and computing structure constants  
symbolic <<;print_:=nil;>> $
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%                              SUPPORT FUNCTIONS                                   %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
procedure kroneckerdelta(k1,k2) $
% Kronecker symbol
begin
  if k1 = k2 then return 1 else return 0 $
end $
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
procedure relieheader $
begin
  write "*****************************************************************" $
  write "ReLie, version 3.0" $ 
  write "A Reduce program for Lie group analysis of differential equations" $
  write "(c) Francesco Oliveri (foliveri@unime.it) - 2021" $
  write "Symmetry, vol. 13, 1826 (1-39), 2021" $
  write "Last update August 9, 2021" $
  write "http://mat521.unime.it/oliveri/" $
  write "*****************************************************************" $
end $
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
procedure bincoeff(n,k) $
% binomial coefficient
begin
  integer k1 $
  return for k1:=0:k-1 product ((n-k1)/(k1+1)) $  
end $
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
procedure letterlist(obj,n) $
% obj is a symbol, n a positive integer
% Builds the list {obj1,...,objn} 
begin
  integer k1 $
  return for k1:=1:n collect mkid(obj,k1) $
end $  
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
procedure listletter(lis,ch) $
% lis is a list, ch a symbol
% Builds the list {part(lis,1)ch,...part(lis,last)ch}
begin
 scalar x $
  return foreach x in lis collect mkid(x,ch) $
end $  
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
procedure letterlistvar(obj,lis) $
% obj is a symbol, lis a list
% Builds the list where each element is obj followed by the generic element of lis
begin
  integer k1 $
  return for k1:=1:length(lis) collect mkid(obj,part(lis,k1)) $
end $  
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
procedure zerolist(n) $
% Returns a list of n zeros, empty list if n<=0
begin
  integer k1 $
  if (n leq 0) then
    return {}
  else
    return for k1:=1:n collect 0 $
end $
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
procedure scalarmult(obj,lis) $
% Returns a list whose i-th element is the product of the scalar obj 
% and the i-th element of list lis
begin
  scalar x $
  return foreach x in lis collect obj*x $
end $
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
procedure scalarproduct(lis1,lis2) $
% Returns the scalar product of two lists with the same number of elements
begin
  integer k1 $
  return for k1:=1:length(lis1) sum part(lis1,k1)*part(lis2,k1) $
end $
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
procedure sumlist(lis1,lis2) $
% Sums element by element two lists with the same number of elements
begin
  integer k1 $
  return for k1:=1:length(lis1) collect part(lis1,k1)+part(lis2,k1) $
end $
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
procedure delzero(lis) $
% Deletes zeros from a list 
begin
  scalar x $
  return
    foreach x in lis join 
      if x neq 0 then {x} else {} $
end $
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
procedure allcoeffs(lis1,lis2) $
% Returns the list of coefficients of lis1 (list of polynomyals) 
% with respect to the variables in list lis2 
  if lis2={} then 
    lis1 
  else 
    allcoeffs(foreach x in lis1 join coeff(x,first lis2),rest lis2) $
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
procedure offprintcrack $
 symbolic <<;print_:=nil;>> $
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
procedure onprintcrack $    
  symbolic <<;print_:=1;>> $
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
procedure combrepnext(temp,n,k) $
begin
  scalar temp1,temp2,temp3 $
  integer k1,k2 $
  temp1:={} $
  temp2:={} $
  k2:=k $
  while (k2>0 and part(temp,k2) = n) do 
    k2:=k2-1 $
  if k2 = 0 then 
    return temp1 
  else
  begin
    temp3:=part(temp,k2)+1 $
    temp1:= append(temp1,{temp3}) $
    for k1:=k2+1:k do 
      temp1:= append(temp1,{temp3}) $
    for k1:=1:k2-1 do 
      temp2:=append(temp2,{part(temp,k1)}) $
  end $  
  temp1:=append(temp2,temp1) $
  return temp1 $
end $
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
procedure combrep(n,k) $
% Combinations with repetition of k elements chosen in {1,2,...,n}  
begin
  scalar temp,temp_all $
  temp_all:={} $
  temp:=for k1:=1:k collect 1 $
  temp_all:=append(temp_all,{temp}) $
  while (length(temp) neq 0) do
  begin
    temp:=combrepnext(temp,n,k) $
    if (length(temp) neq 0) then temp_all:=append(temp_all,{temp}) $
  end $
  return temp_all $
end $
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
procedure combnorepinit(n,k) $
begin
  scalar temp1 $
  integer i1 $
  array temparr(k-1);
  for i1:=0:k-1 do temparr(i1):=i1;
  temp1:=for i1:=0:k-1 collect temparr(i1)+1 $
  clear temparr $
  return temp1 $
end $
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
procedure combnorepnext(temp,n,k) $
begin
  scalar temp1 $
  integer j,j1 $
  array temparr(k-1) $
  for i1:=0:k-1 do temparr(i1):=part(temp,i1+1)-1 $
  j:=k-1;
  temparr(j):=temparr(j)+1 $
  while ((j > 0) and (temparr(j) >= n - k + 1 + j)) do
  begin
  j:=j-1 $
  temparr(j):=temparr(j)+1 $
  end $
  if (temparr(0) > n - k) then  % Combination (n-k, n-k+1, ..., n) reached 
    temp1:={}                   % No more combinations can be generated
  else
  begin
    for j1:=j+1:k-1 do
    temparr(j1):=temparr(j1-1) + 1 $
    temp1:=for j1:=0:k-1 collect temparr(j1)+1 $
  end $   
  clear temparr $
  return temp1 $
end $
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
procedure combnorep(n,k) $
% Combinations without repetition of k elements chosen in {1,2,...,n}  
begin
  scalar temp,temp_all $
  temp_all:={} $
  temp:=combnorepinit(n,k) $
  temp_all:=append(temp_all,{temp}) $
  while (length(temp) neq 0) do
  begin
    temp:=combnorepnext(temp,n,k) $
    if (length(temp) neq 0) then temp_all:=append(temp_all,{temp}) $
  end $
  return temp_all $
end $
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
procedure membership(elem,lis) $
% Returns the number of occurrences of elem in list lis
begin
  scalar x,temp $
  temp:=foreach x in lis join
    if x=elem then {x} else {} $
  return length(temp) $  
end $
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
procedure removeelement(lis,elem) $
% Removes elem from lis
begin
  scalar x $
  return 
    foreach x in lis join
      if x neq elem then {x} else {} $
end $
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
procedure removemultiple(lis1,lis2) $
% Removes from lis1 the elements belonging to lis2
begin
  scalar elem $
  return 
    foreach elem in lis1 join
      if membership(elem,lis2)>0 then {} else {elem} $
end $
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
procedure counting(lis,n) $
% lis is a list of positive integers, n its maximum 
% Returns a list of n numbers where the k-th element is the number of values k in lis
begin
  scalar temp $
  integer k1, x $
  array temparr(n) $
  foreach x in lis do temparr(x):=temparr(x)+1 $
  temp:=for k1:=1:n collect temparr(k1) $
  clear temparr $
  return temp $  
end $
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
procedure newordering(lis,ind) $
% lis is a list and ind a list of distinct integers in the range [1..length(lis)]
% returns a list containing the elements of lis according to the indexex in ind
begin
  scalar temp $
  integer k1,ok $
  temp:=lis $
  if length(lis) neq length(ind) then
    write "Arguments are lists of different length!" 
  else
  begin
    ok:=0 $
    for k1:=1:length(ind) do
      if membership(k1,ind)=1 then ok:=ok+1 $
    if ok=length(ind) then
      temp:=foreach k1 in ind collect part(lis,k1)
    else
      write "The second list contains equal integers or integers not in the right range!" $
  end $
  return temp $
end $
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
procedure dependence(lis1,lis2) $
% Declares the dependence of the elements in lis1 upon all the elements in lis2 
begin
  scalar y,x $
  foreach y in lis1 do
    foreach x in lis2 do
      depend y,x $
end $
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
procedure nodependence(lis1,lis2) $
% Removes the dependence of the elements in lis1 upon all the elements in lis2 
begin
  scalar y,x $
  foreach y in lis1 do
    foreach x in lis2 do
      nodepend y,x $
end $
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
procedure listdf(lis,var,n) $
begin
% Returns a list containing the n-order derivatives of the elements of lis 
% w.r.t. var
  scalar y $
  return 
    foreach y in lis collect df(y,var,n) $
end $
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
procedure essentialpars(gens,vars) $
% Takes a list of generators of a Lie multiparameter group and drops
% the generators corresponding to non essential parameters.
begin
  integer k1,k2 $
  scalar res,temp,cc,sol,ccsol $
  symbolic <<;print_:=nil;>> $
  res:={part(gens,1)} $
  for k1:=2:length(gens) do
  begin
    temp:=zerolist(length(vars)) $
    for k2:=1:length(res) do
    begin
      cc:=letterlist(c,length(res)+1) $
      temp:=sumlist(temp,scalarmult(part(cc,k2),part(res,k2))) $
    end $
    temp:=sumlist(temp,scalarmult(part(cc,length(res)+1),part(gens,k1))) $
    sol:=crack(temp,{},cc,vars) $
    ccsol:=delzero(sub(part(part(sol,1),2),cc)) $
    if (ccsol={}) then
      res:=append(res,{part(gens,k1)}) $ 
  end $
  symbolic <<;print_:=nil; print_all:=t>> $
  return res $
end $
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%     MAIN FUNCTIONS
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
procedure reliecheck $
% This function performs some simple checks on the user input 
begin
  integer ok $
  okinputrelie:={} $
  ok:=0 $
  % 1 Check jetorder, xvar, uvar
  if (fixp(jetorder) and jetorder>0) then 
    ok:=ok+1  
  else
    write "A positive integer value for 'jetorder' is needed!" $
  if (depth(xvar)=1 and length(xvar)>0) then 
    ok:=ok+1  
  else
    write "The list 'xvar' of independent variable(s) is needed!" $
  if (depth(uvar)=1 and length(uvar)>0) then 
    ok:=ok+1  
  else
    write "The list 'uvar' of dependent variable(s) is needed!" $
  okinputrelie:= append(okinputrelie,{ok}) $  
  %2 Check for equivalence transformations  
  if (arborder neq -1) then
  begin
    if (zorder leq jetorder) then
    begin 
      if (depth(arbelem) and length(arbelem)>0) then 
        okinputrelie:= append(okinputrelie,{1})  
      else
      begin
        okinputrelie:= append(okinputrelie,{0}) $
        write "The list 'arbelem' of arbitrary elements for equivalence transformations is needed!" $
      end $
    end 
    else
    begin
      write "zorder cannot be greater than jetorder." $
      okinputrelie:= append(okinputrelie,{0}) $
    end $
  end   
  else
    okinputrelie:= append(okinputrelie,{0}) $   
  %3 Check for conditional symmetries
  if (nonclassical > 0 and nonclassical leq length(xvar)) then 
  begin
    if (depth(qcond)=1 and length(qcond)>0) then
    begin  
      if length(qcond) leq length(uvar) and max(qcond) leq length(uvar) then
        okinputrelie:= append(okinputrelie,{1})
      else
      begin
        write "The list 'qcond' for conditional symmetries is not compatible!" $  
        okinputrelie:= append(okinputrelie,{0}) $
      end $  
    end
    else  
    begin
      write "The list 'qcond' for conditional symmetries is needed!" $
      okinputrelie:= append(okinputrelie,{0}) $
    end $
  end 
  else
    okinputrelie:= append(okinputrelie,{0}) $  
  %4 Check for contact transformations
  if contact = 1 then
  begin
    if (depth(uvar)=1 and length(uvar) = 1) then
      okinputrelie:= append(okinputrelie,{1})
    else
    begin
      write "No true contact symmetries for systems of differential equations!" $
      okinputrelie:= append(okinputrelie,{0}) $
    end     
  end
  else 
    okinputrelie:= append(okinputrelie,{0}) $  
  %5 Check for approximate symmetries
  if approxorder > 0 then
  begin
    if epsilon^(approxorder+1) = 0 then
      okinputrelie:= append(okinputrelie,{1})
    else
    begin
      write "the rule 'let epsilon^(approxorder+1)=0' is missing!" $
      okinputrelie:= append(okinputrelie,{0}) $
    end     
  end
  else 
    okinputrelie:= append(okinputrelie,{0}) $  
  %6 Check diffeqs,leadders and solvability or lagrangian
  if variational = 0 then 
  begin
    ok:=0 $
    if (depth(diffeqs)=1 and length(diffeqs)>0) then 
      ok:=ok+1 
    else
      write "The list 'diffeqs' of differential equation(s) is missing!" $ 
    if (depth(leadders)=1 and length(leadders)>0) then 
      ok:=ok+1  
    else
      write "The list 'leadders' of leading derivative(s) is missing!" $ 
    if ok = 2 then 
    begin
      if length(diffeqs)=length(leadders) then
        ok:=ok+1 
      else 
        write "Lists of differential equations and leading derivatives are of different length!" $ 
    end $
    if ok < 3 then    
    begin  
      write "Check 'diffeqs' and/or 'leadders'!" $  
      okinputrelie:= append(okinputrelie,{0}) $
    end 
    else
      okinputrelie:= append(okinputrelie,{3}) $  
  end
  else 
  begin
    if (depth(lagrangian)=1 and length(lagrangian)=1) then
      okinputrelie:= append(okinputrelie,{7}) 
    else
    begin
      write "The Lagrangian is missing!" $
      okinputrelie:=append(okinputrelie,{0}) $
    end $
  end $   
end $   
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
procedure reliestatus $
begin
  integer relienumber $
  relienumber:=
     for k1:=1:length(okinputrelie) sum part(okinputrelie,k1)*10^(k1-1) $
  if membership(relienumber,{3,13,103,1003,10003,10103,11003})>0 then 
    write "You may only call relieprol() or generatealgebra(k) (k=1,2,3)." 
  else if relienumber=300003 then
    write "You may compute exact point symmetries." 
  else if relienumber=310003 then
    write "You may compute approximate point symmetries."   
  else if relienumber=300013 then
    write "You may compute equivalence transformations." 
  else if relienumber=300103 then
    write "You may compute exact conditional symmetries." 
  else if relienumber=310103 then
    write "You may compute approximate conditional symmetries." 
  else if relienumber=301003 then
    write "You may compute exact contact symmetries." 
  else if relienumber=311003 then
    write "You may compute approximate contact symmetries." 
  else if relienumber=700003 then
    write "You may compute exact variational symmetries."
  else if relienumber=710003 then
    write "You may compute approximate variational symmetries."
  else
    write "Check your input!" $   
end $
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
procedure init_infinitesimals_exact $
begin
  integer k1,njet1 $
  scalar xuvar,xuuxvar,uxvar,infinitesimals_mu $
  xuvar:=append(xvar,uvar) $
  temp_uvar:=letterlistvar(temp_,uvar) $
  dependence(temp_uvar,xvar) $
  jet:=jet_space_exact() $
  if part(okinputrelie,4) = 0 then  % Point symmetries
  begin
    allinfinitesimals:=append(letterlistvar(xi_,xvar),letterlistvar(eta_,uvar)) $
    dependence(allinfinitesimals,xuvar) $
    if part(okinputrelie,2)=1 then   % equivalence transformations
    begin
      zvar:=init_zvar() $
      temp_arbelem:=letterlistvar(temp_,arbelem) $
      dependence(temp_arbelem,part(zvar,2)) $
      infinitesimals_mu:=letterlistvar(mu_,arbelem) $
      dependence(infinitesimals_mu,append(xuvar,arbelem)) $
      jet_equiv:=jet_space_equivalence_exact() $
      if generalequiv = 1 then
        dependence(allinfinitesimals,arbelem) $
      allinfinitesimals:=append(allinfinitesimals,infinitesimals_mu) $ 
    end $
    allinfinitesimals:={allinfinitesimals,allinfinitesimals} $
  end 
  else   % Contact symmetries
  begin
    njet1:=2*length(xvar)+1 $
    xuuxvar:=for k1:=1:njet1 collect part(part(jet,3),k1) $
    dependence({omega},xuuxvar) $
    uxvar:=for k1:=length(xvar)+2:njet1 collect part(xuuxvar,k1) $
    allinfinitesimals:=
    for k1:=1:length(xvar) collect (-df(omega,part(uxvar,k1))) $
    temp:=for k1:=1:length(xvar) sum 
          part(uxvar,k1)*df(omega,part(uxvar,k1)) $  
    allinfinitesimals:=append(allinfinitesimals,{omega-temp}) $
    for k1:=1:length(xvar) do
      allinfinitesimals:=append(allinfinitesimals,
             {df(omega,part(xvar,k1))+part(uxvar,k1)*df(omega,part(uvar,1))}) $ 
    allinfinitesimals:={allinfinitesimals,{omega}} $         
  end $
end $
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
procedure init_infinitesimals_approximate $
begin 
  integer k1,k2,k3,k4,njet1,njet,n_perturb $
  scalar kk,uvar0,xuvar,allxiterms,alletaterms,temp,expansion_xi,expansion_eta,
         term,temp_xi,temp_eta,perturb,combinations,xuuxvar,uuxvar,
         allomegaterms,expansion_omega,temp_omega $
  uvarall:=for k1:=0:approxorder collect listletter(uvar,k1) $      
  uvar0:=part(uvarall,1) $
  xuvar:=append(xvar,uvar0) $
  temp_uvar:=letterlistvar(temp_,uvar) $
  dependence(temp_uvar,xvar) $
  temp_uvarall:=
    for k1:=1:approxorder+1 collect letterlistvar(temp_,part(uvarall,k1)) $
  for k1:=1:approxorder+1 do
    dependence(part(temp_uvarall,k1),xvar) $
  jet:=jet_space_exact() $
  jetapprox:=jet_space_approx() $
  jetsplit:=jet_space_split_approx() $
  expansion_temp_uvar:=part(temp_uvarall,1) $
  for k1:=1:approxorder do
    expansion_temp_uvar:=sumlist(expansion_temp_uvar,
                         scalarmult(epsilon^k1,part(temp_uvarall,k1+1))) $
  if part(okinputrelie,4) = 0 then  % Point symmetries 
  begin
    allxiterms:={} $
    alletaterms:={} $
    for k1:=0:approxorder do
    begin
      temp:=letterlistvar(mkid(mkid(xi,k1),_),xvar) $
      dependence(temp,xuvar) $
      allxiterms:=append(allxiterms,{temp}) $
      temp:=letterlistvar(mkid(mkid(eta,k1),_),uvar) $
      dependence(temp,xuvar) $
      alletaterms:=append(alletaterms,{temp}) $
    end $
    perturb:=zerolist(length(uvar)) $
    for k1:=1:approxorder do
      perturb:=sumlist(perturb,scalarmult(epsilon^k1,part(uvarall,k1+1))) $
    perturb:=append({epsilon},perturb) $
    n_perturb:=length(perturb) $
    expansion_xi:=part(allxiterms,1) $
    expansion_eta:=part(alletaterms,1) $
    for k1:=1:approxorder do
    begin
      combinations:=combrep(n_perturb,k1) $
      for k2:=1:length(combinations) do
      begin
        kk:=counting(part(combinations,k2),n_perturb) $
        term:=1 $
        for k3:=1:length(kk) do
        begin
          k4:=part(kk,k3) $
          term:=term*(part(perturb,k3))^k4/factorial(k4) $
        end $  
        temp_xi:=part(allxiterms,part(kk,1)+1) $
        temp_eta:=part(alletaterms,part(kk,1)+1) $
        for k3:=2:length(kk) do
        begin
          if part(kk,k3) neq 0 then
          begin
            temp_xi:=listdf(temp_xi,part(uvar0,k3-1),part(kk,k3)) $
            temp_eta:=listdf(temp_eta,part(uvar0,k3-1),part(kk,k3)) $
          end $
        end $
        expansion_xi:=sumlist(expansion_xi,scalarmult(term,temp_xi)) $
        expansion_eta:=sumlist(expansion_eta,scalarmult(term,temp_eta)) $
      end $
    end $ 
    allinfinitesimals:={} $
    for k1:=1:approxorder+1 do
      allinfinitesimals:=append(allinfinitesimals,part(allxiterms,k1)) $
    for k1:=1:approxorder+1 do
      allinfinitesimals:=append(allinfinitesimals,part(alletaterms,k1)) $
    allinfinitesimals:={append(expansion_xi,expansion_eta),allinfinitesimals} $  
  end 
  else   % Contact symmetries
  begin
    njet1:=2*length(xvar)+1 $
    njet:=length(xvar)
          +bincoeff(length(xvar)+jetorder,jetorder) $
    xuuxvar:=for k1:=length(xvar)+1:njet1 collect part(part(jetapprox,2),k1) $
    allomegaterms:=for k1:=0:approxorder collect {mkid(omega,k1)} $
    dependence(allomegaterms,xuuxvar) $
    uuxvar:=removemultiple(xuuxvar,xvar) $
    n_perturb:=length(xvar)+1 $
    perturb:=zerolist(n_perturb) $
    for k1:=1:approxorder do
    begin
      temp:=for k2:=k1*njet+1:k1*njet+n_perturb collect part(part(jetapprox,2),k2) $
      perturb:=sumlist(perturb,scalarmult(epsilon^k1,temp)) $
    end $           
    perturb:=append({epsilon},perturb) $  
    n_perturb:=n_perturb+1 $
    expansion_omega:=part(allomegaterms,1) $
    for k1:=1:approxorder do
    begin
      combinations:=combrep(n_perturb,k1) $
      for k2:=1:length(combinations) do
      begin
        kk:=counting(part(combinations,k2),n_perturb) $
        term:=1 $
        for k3:=1:length(kk) do
        begin
          k4:=part(kk,k3) $
          term:=term*(part(perturb,k3))^k4/factorial(k4) $
        end $  
        temp_omega:=part(allomegaterms,part(kk,1)+1) $
        for k3:=2:length(kk) do
        begin
          if part(kk,k3) neq 0 then
          begin
            temp_omega:=listdf(temp_omega,part(uuxvar,k3-1),part(kk,k3)) $
          end $
        end $
        expansion_omega:=sumlist(expansion_omega,scalarmult(term,temp_omega)) $
      end $
    end $
    omega:=part(expansion_omega,1) $
    allinfinitesimals:=for k1:=1:length(xvar) collect
      (-df(omega,part(uuxvar,k1+1))) $
    temp:=for k1:=1:length(xvar) sum 
      df(part(expansion_temp_uvar,1),
        part(xvar,k1))*df(omega,part(uuxvar,k1+1)) $  
    allinfinitesimals:=append(allinfinitesimals,{omega-temp}) $
    for k1:=1:length(xvar) do
      allinfinitesimals:=append(allinfinitesimals,
        {df(omega,part(xvar,k1))+df(part(expansion_temp_uvar,1),part(xvar,k1))
                                *df(omega,part(uvar0,1))}) $  
    allinfinitesimals:=arrangeapprox(allinfinitesimals) $
    allomegaterms:=for k1:=1:approxorder+1 collect part(part(allomegaterms,k1),1) $
    allinfinitesimals:={allinfinitesimals,allomegaterms} $
  end $
end $
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
procedure relieinit $
% This function performs additional checks on the user input and sets
% variables and functions necessary to do all computations
begin
  scalar vars,temp,allphiterms,expansion_phi,term,combinations,
         temp_phi,perturb $
  integer ok,k1,k2,k3,k4,njetm1,njet,n_perturb $
  reliecheck() $
  if part(okinputrelie,1) = 3 then   % Minimal input
  begin
    if part(okinputrelie,5) = 0 then  % Exact symmetries
      init_infinitesimals_exact() 
    else  % Approximate symmetries
      init_infinitesimals_approximate() $
    if part(okinputrelie,6) = 3 then    % Symmetries of differential equations
    begin
      allphi:={{},{}} $
      ok := checksolvability() $
      if ok neq 1 then    
      begin  
        write "Differential equations cannot be solved w.r.t. leading derivatives!" $  
        okinputrelie:= for k1:=1:5 collect part(okinputrelie,k1) $
        okinputrelie:=append(okinputrelie,{0}) $
      end $
    end 
    else if part(okinputrelie,6)=7 then   % variational symmetries
    begin
      if part(okinputrelie,5)=0  then     % exact variational symmetries
      begin
        njetm1:=length(xvar)+length(uvar)*
           bincoeff(length(xvar)+jetorder-1,jetorder-1) $
        allphi:=letterlistvar(phi_,xvar) $
        vars:=for k1:=1:njetm1 collect part(part(jet,3),k1) $   
        dependence(allphi,vars) $
        allphi:={allphi,allphi} $
      end 
      else    % approximate variational symmetries
      begin
        njetm1:=length(xvar)+length(uvar)*
           bincoeff(length(xvar)+jetorder-1,jetorder-1) $
        njet:=length(xvar)+length(uvar)*
           bincoeff(length(xvar)+jetorder,jetorder) $   
        nincjet:=length(uvar)*
           bincoeff(length(xvar)+2*jetorder,2*jetorder) $     
        vars:=for k1:=1:njetm1 collect part(part(jetapprox,2),k1) $
        allphiterms:={} $
        for k1:=0:approxorder do
        begin
          temp:=letterlistvar(mkid(mkid(phi,k1),_),xvar) $
          dependence(temp,vars) $
          allphiterms:=append(allphiterms,{temp}) $
        end $
        vars:=removemultiple(vars,xvar) $
        njetm1:=njetm1-length(xvar) $
        perturb:=zerolist(njetm1) $
        for k1:=1:approxorder do
        begin
          temp:=for k2:=length(xvar)+1+k1*nincjet:length(xvar)+k1*nincjet+njetm1 
                collect part(part(jetapprox,2),k2) $
          perturb:=sumlist(perturb,
                   scalarmult(epsilon^k1,temp)) $
        end $     
        perturb:=append({epsilon},perturb) $  
        n_perturb:=njetm1+1 $
        expansion_phi:=part(allphiterms,1) $
        for k1:=1:approxorder do
        begin
          combinations:=combrep(n_perturb,k1) $
          for k2:=1:length(combinations) do
          begin
            kk:=counting(part(combinations,k2),n_perturb) $
            term:=1 $
            for k3:=1:length(kk) do
            begin
              k4:=part(kk,k3) $
              term:=term*(part(perturb,k3))^k4/factorial(k4) $
            end $  
            temp_phi:=part(allphiterms,part(kk,1)+1) $
            for k3:=2:length(kk) do
            begin
              if part(kk,k3) neq 0 then
              begin
                temp_phi:=listdf(temp_phi,part(vars,k3-1),part(kk,k3)) $
              end $
            end $
            expansion_phi:=sumlist(expansion_phi,scalarmult(term,temp_phi)) $
          end $
        end $ 
        allphi:={} $
        for k1:=1:length(allphiterms) do
          allphi:=append(allphi,part(allphiterms,k1)) $
        allphi:={expansion_phi,allphi} $ 
        expand_lagrangian() $     
      end $   
    end $  
    steprelie:=1 $
    reliestatus() $
  end $
end $  
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
procedure checksolvability $  
begin
  integer ok,k1,k2,k3,k4 $
  scalar alldiffeqs,allleadders,newder,tempder,neweqn,tempeqn,combinations,
         tempcomb,substconditional,tempdiffeqs,temp1,temp2 $
  alldiffeqs:=diffeqs $
  allleadders:=leadders $
  if part(okinputrelie,5) = 0 then   % exact symmetries  
  begin
    if part(okinputrelie,3) = 1 then % conditional symmetries
    begin
      substconditional:={} $
      for k1:=1:nonclassical-1 do
        substconditional:=append(substconditional,
                          {part(part(allinfinitesimals,2),k1)=0}) $
      substconditional:=append(substconditional,
                          {part(part(allinfinitesimals,2),nonclassical)=1}) $ 
      allinfinitesimals:=sub(substconditional,allinfinitesimals) $
      for k1:=1:length(qcond) do
      begin
        newder:=df(part(temp_uvar,part(qcond,k1)),part(xvar,nonclassical)) $
        allleadders:=append(allleadders,{newder}) $
        neweqn:=for k2:=1:length(xvar) sum
          part(part(allinfinitesimals,2),k2)*
               df(part(temp_uvar,part(qcond,k1)),part(xvar,k2)) $
        neweqn:=neweqn-part(part(allinfinitesimals,2),length(xvar)+part(qcond,k1)) $
        alldiffeqs:=append(alldiffeqs,{neweqn}) $
        for k2:=1:jetorder-1 do
        begin
          combinations:=combrep(length(xvar),k2) $
          for k3:=1:length(combinations) do
          begin
            tempeqn:=neweqn $
            tempder:=newder $
            for k4:=1:length(part(combinations,k3)) do
            begin
              tempeqn:=dlie(tempeqn,part(xvar,part(part(combinations,k3),k4))) $
              tempder:=df(tempder,part(xvar,part(part(combinations,k3),k4))) $
            end $
            alldiffeqs:=append(alldiffeqs,{tempeqn}) $  
            allleadders:=append(allleadders,{tempder}) $
          end $
        end $  
      end $  
      alldiffeqs:=arrange(alldiffeqs) $
      allleadders:=arrange(allleadders) $
    end $
  end  
  else                                 % approximate symmetries
  begin
    if part(okinputrelie,3) = 1 then % conditional symmetries
    begin
      substconditional:={} $
      for k1:=1:nonclassical-1 do
      begin
        for k2:=0:approxorder do
          substconditional:=append(substconditional,
               {part(part(allinfinitesimals,2),k2*length(xvar)+k1)=0}) $
      end $  
      substconditional:=append(substconditional,
          {part(part(allinfinitesimals,2),nonclassical)=1}) $
      for k1:=1:approxorder do
        substconditional:=append(substconditional,
          {part(part(allinfinitesimals,2),k1*length(xvar)+nonclassical)=0}) $
      allinfinitesimals:=sub(substconditional,allinfinitesimals) $
      for k1:=1:length(qcond) do
      begin
        newder:=df(part(temp_uvar,part(qcond,k1)),part(xvar,nonclassical)) $
        allleadders:=append(allleadders,{newder}) $
        neweqn:=for k2:=1:length(xvar) sum
          part(part(allinfinitesimals,1),k2)*df(part(temp_uvar,part(qcond,k1)),part(xvar,k2)) $
        neweqn:=neweqn-part(part(allinfinitesimals,1),length(xvar)+part(qcond,k1)) $
        alldiffeqs:=append(alldiffeqs,{neweqn}) $
        for k2:=1:jetorder-1 do
        begin
          combinations:=combrep(length(xvar),k2) $
          for k3:=1:length(combinations) do
          begin
            tempeqn:=neweqn $
            tempder:=newder $
            tempcomb:=part(combinations,k3) $
            for k4:=1:length(tempcomb) do
            begin
              tempeqn:=dlieapprox(tempeqn,part(xvar,part(tempcomb,k4))) $
              tempder:=df(tempder,part(xvar,part(tempcomb,k4)))$
            end $  
            alldiffeqs:=append(alldiffeqs,{tempeqn}) $  
            allleadders:=append(allleadders,{tempder}) $
          end $
        end $    
      end $
    end $  
    alldiffeqs:=inversearrange(alldiffeqs) $
    alldiffeqs:=substitute_expansion_uvar(alldiffeqs) $
    tempdiffeqs:={} $
    for k1:=1:length(alldiffeqs) do
    begin
      temp1:=part(alldiffeqs,k1) $
      temp2:=sub(epsilon=0,temp1) $
      temp2:=temp2 +
        for k2:=1:approxorder sum
          epsilon^k2*sub(epsilon=0,df(temp1,epsilon,k2))/factorial(k2) $
      tempdiffeqs:=append(tempdiffeqs,{temp2}) $    
    end $
    alldiffeqs:=arrangeapprox(tempdiffeqs) $
    allleadders:=inversearrange(allleadders) $
    allleadders:=substitute_expansion_uvar(allleadders) $
    allleadders:=arrangeapprox(allleadders) $
    alldiffeqs:=allcoeffs(alldiffeqs,{epsilon}) $
    allleadders:=allcoeffs(allleadders,{epsilon}) $  
  end $  
  solutiondedv:=solve(alldiffeqs,allleadders) $
  if length(solutiondedv)=1 then
  begin
    solutiondedv:=part(solutiondedv,1) $
    ok:=1 $ 
  end  
  else
    ok:=0 $
  return ok $     
end $
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
procedure expand_lagrangian $
begin
  scalar temp,temp1 $
  integer k1 $
  lagrangian:=inversearrange(lagrangian) $
  lagrangian:=substitute_expansion_uvar(lagrangian) $
  lagrangian:=part(arrangeapprox(lagrangian),1) $
  lagrangian:=sub(epsilon=nolispe,lagrangian) $
  temp:={sub(nolispe=0,lagrangian)} $
  for k1:=1:approxorder do
  begin
    temp1:=df(lagrangian,nolispe,k1)/factorial(k1) $
    temp:=append(temp,{sub(nolispe=0,temp1)}) $
  end $
  lagrangian:=0 $
  for k1:=1:length(temp) do
    lagrangian:=lagrangian+epsilon^(k1-1)*part(temp,k1) $
  lagrangian:={lagrangian} $    
end $
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
procedure jet_space_exact $
% ***************************************************************************
% ** this procedure creates a list of the jetorder - jet space
% ***************************************************************************
begin
  scalar internalders,trueders,aninternalder,atrueder,
         indices,ind,combinations $
  integer k1,k2,k3,k4,incjetorder $
  internalders:=append(xvar,uvar) $
  trueders:=append(xvar,temp_uvar) $
  indices:={} $
  for k1:=1:length(xvar) do
    indices:=append(indices,{{k1,{-1}}}) $
  for k1:=1:length(uvar) do
    indices:=append(indices,{{k1,{0}}}) $
  if (part(okinputrelie,6) = 7) then
    incjetorder:=2*jetorder
    %incjetorder:=jetorder 
  else
    incjetorder:=jetorder $    
  for k1:=1:incjetorder do
  begin
    combinations:=combrep(length(xvar),k1) $
    for k2:=1:length(uvar) do
    begin
      for k3:=1:length(combinations) do
      begin
        aninternalder:=mkid(part(uvar,k2),_) $
        atrueder:=part(temp_uvar,k2) $
        ind:={k2} $
        for k4:=1:k1 do
        begin
          aninternalder:=mkid(aninternalder,part(xvar,part(part(combinations,k3),k4))) $
          atrueder:=df(atrueder,part(xvar,part(part(combinations,k3),k4))) $
        end $
        ind:=append(ind,{part(combinations,k3)}) $
        indices:=append(indices,{ind}) $
        internalders:=append(internalders,{aninternalder}) $
        trueders:=append(trueders,{atrueder}) $
      end $
    end $
  end $  
  return {indices,trueders,internalders} $
end $
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
procedure jet_space_approx $
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% this procedure creates a list of the jetorder -- jet approximate space
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
begin
  scalar internalders,trueders,aninternalder,atrueder,
         indices,ind,combinations $ 
  integer k0,k1,k2,k3,k4,incjetorder $
  internalders:=xvar $
  trueders:=xvar $
  for k0:=1:approxorder+1 do
  begin
    for k2:=1:length(uvar) do
    begin
      internalders:=append(internalders,{part(part(uvarall,k0),k2)}) $
      trueders:=append(trueders,{part(part(temp_uvarall,k0),k2)}) $
    end $
    if (part(okinputrelie,6) = 7) then
      incjetorder:=2*jetorder
      %incjetorder:=jetorder 
    else
      incjetorder:=jetorder $    
    for k1:=1:incjetorder do
    begin
      combinations:=combrep(length(xvar),k1) $
      for k2:=1:length(uvar) do
      begin
        for k3:=1:length(combinations) do
        begin
          aninternalder:=mkid(part(part(uvarall,k0),k2),_) $
          atrueder:=part(part(temp_uvarall,k0),k2) $
          for k4:=1:k1 do
          begin
            aninternalder:=mkid(aninternalder,part(xvar,part(part(combinations,k3),k4))) $
            atrueder:=df(atrueder,part(xvar,part(part(combinations,k3),k4))) $
          end $
          internalders:=append(internalders,{aninternalder}) $
          trueders:=append(trueders,{atrueder}) $
        end $
      end $
    end $  
  end $
  return {trueders,internalders} $
end $
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
procedure jet_space_split_approx $
% ****************************************************************************
% ** creates a list of the jetorder -- jet space of zeroth dependent variables
% ****************************************************************************
begin
  scalar internalders,aninternalder,combinations,ind,indices $
  integer k1,k2,k3,k4,incjetorder $
  internalders:=xvar $
  indices:={} $
  for k1:=1:length(xvar) do
    indices:=append(indices,{{k1,{-1}}}) $
  for k1:=1:length(uvar) do
  begin
    internalders:=append(internalders,{mkid(part(uvar,k1),0)}) $
    indices:=append(indices,{{k1,{0}}}) $
  end $
  if (part(okinputrelie,6) = 7) then
    incjetorder:=2*jetorder
  else
    incjetorder:=jetorder $    
  for k1:=1:incjetorder do
  begin
    combinations:=combrep(length(xvar),k1) $
    for k2:=1:length(uvar) do
    begin
      for k3:=1:length(combinations) do
      begin
        aninternalder:=mkid(mkid(part(uvar,k2),0),_) $
        ind:={k2} $
        for k4:=1:k1 do
        begin
          aninternalder:=mkid(aninternalder,part(xvar,part(part(combinations,k3),k4))) $
        end $
        ind:=append(ind,{part(combinations,k3)}) $
        indices:=append(indices,{ind}) $
        internalders:=append(internalders,{aninternalder}) $
      end $
    end $
  end $  
  return {indices,internalders} $
end $
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
procedure init_zvar $
begin
% Generation of z-variables (for equivalence transformations)
  scalar indices, internalders $ 
  integer k1,n $
  n:=length(xvar)+length(uvar)*bincoeff(length(xvar)+zorder,zorder) $
  indices:={} $
  internalders:={} $
  for k1:=1:n do
  begin
    indices:=append(indices,{part(part(jet,1),k1)}) $
    internalders:=append(internalders,{part(part(jet,3),k1)}) $
  end $
  return {indices,internalders} $
end $
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
procedure jet_space_equivalence_exact $
% Generation of jet space for equivalent transformations
begin    
  scalar internalders,trueders,indices,temp_zvar,combinations,indices,
         ind,aninternalder,atrueder $ 
  integer k1,k2,k3,k4 $
  internalders:={} $
  trueders:={} $
  indices:={} $
  temp_zvar:=part(zvar,2) $
  for k1:=1:length(arbelem) do   
  begin
    internalders:=append(internalders,{part(arbelem,k1)}) $
    trueders:=append(trueders,{part(temp_arbelem,k1)}) $
    indices:=append(indices,{{k1,{0}}}) $
  end $
  for k1:=1:arborder do
  begin
    combinations:=combrep(length(temp_zvar),k1) $
    for k2:=1:length(arbelem) do
    begin
      for k3:=1:length(combinations) do
      begin
        aninternalder:=mkid(part(arbelem,k2),_) $
        atrueder:=part(temp_arbelem,k2) $
        ind:={k2} $
        for k4:=1:k1 do
        begin
          aninternalder:=mkid(aninternalder,part(temp_zvar,part(part(combinations,k3),k4))) $
          atrueder:=df(atrueder,part(temp_zvar,part(part(combinations,k3),k4))) $
        end $
        ind:=append(ind,{part(combinations,k3)}) $
        indices:=append(indices,{ind}) $
        internalders:=append(internalders,{aninternalder}) $
        trueders:=append(trueders,{atrueder}) $
      end $
    end $
  end $  
  return {indices,trueders,internalders} $
end $    
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
procedure arrange(obj) $
% substitutes true derivatives with internal representations of derivatives
begin
  scalar temp,subst $ 
  integer k1,njet $
  njet:=length(part(jet,1)) $
  subst:=for k1:=1:njet collect
    part(part(jet,2),k1)=part(part(jet,3),k1) $
  temp:=sub(subst,obj) $
  if part(okinputrelie,2)=1 then 
  begin   
    njet:=length(part(jet_equiv,1)) $
    subst:=for k1:=1:njet collect
      part(part(jet_equiv,2),k1)=part(part(jet_equiv,3),k1) $   
    temp:=sub(subst,temp) $
  end $
  return temp $  
end $
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
procedure inversearrange(obj) $
% substitutes internal representations of derivatives with true derivatives 
begin
  scalar subst,temp $
  integer k1,njet $
  temp:=obj $
  njet:=length(part(jet,1)) $
  subst:=for k1:=1:njet collect
    part(part(jet,3),k1)=part(part(jet,2),k1) $
  temp:=sub(subst,obj) $  
  if part(okinputrelie,2)=1 then 
  begin   
    njet:=length(part(jet_equiv,1)) $
    for k1:=1:njet collect
      part(part(jet_equiv,3),k1)=part(part(jet_equiv,2),k1) $
    temp:=sub(subst,temp) $   
  end $
  return temp $  
end $
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
procedure arrangeapprox(obj) $
% substitutes true derivatives with internal representations of derivatives
begin
  scalar temp $
  integer k1 $
  temp:=for k1:=1:length(part(jetapprox,1)) collect
    part(part(jetapprox,1),k1)=part(part(jetapprox,2),k1) $  
  return sub(temp,obj) $   
end $
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
procedure substitute_expansion_uvar(obj) $
% substitutes expansions of dependent variables
begin
  scalar temp $
  integer k1 $
  temp:=for k1:=1:length(uvar) collect
    part(temp_uvar,k1)=part(expansion_temp_uvar,k1) $
  return sub(temp,obj) $
end $
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
procedure dlie(obj,var) $
% **********************************
% ** operator of Lie derivative   **
% **********************************
begin
  integer k1,njet $
  scalar temp $
  njet:=length(xvar)+length(uvar)*bincoeff(length(xvar)+jetorder,jetorder) $
  temp:=df(obj,var)+
    for k1:=length(xvar)+1:njet sum
      df(obj,part(part(jet,3),k1))*df(part(part(jet,2),k1),var) $
  return arrange(temp) $    
end $
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
procedure dlieapprox(obj,var) $
% **********************************************
% ** operator of approximate Lie derivative   **
% **********************************************
begin
  integer k1,njet $
  scalar temp $
  njet:=length(part(jetapprox,1)) $
  temp:=df(obj,var)+
    for k1:=length(xvar)+1:njet sum
      df(obj,part(part(jetapprox,2),k1))*df(part(part(jetapprox,1),k1),var) $ 
  return arrangeapprox(temp) $
end $
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
procedure dliestar(obj,var) $
% ****************************************************************
% ** operator of Lie derivative for equivalence transformations **
% ****************************************************************
begin
  integer k1 $
  scalar temp $
  temp:= 
    df(obj,var)+ 
    for k1:=1:length(arbelem) sum
      df(obj,part(arbelem,k1))*df(part(temp_arbelem,k1),var) $
  return arrange(temp) $ 
end $
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
procedure etax(a,lis) $
% Prolonged infinitesimal generator for Lie point symmetries
begin
  scalar der,res,temp $
  integer k1,k2 $
  if lis={-1} then
    res:=part(part(allinfinitesimals,1),a)
  else if lis={0} then
    res:=part(part(allinfinitesimals,1),length(xvar)+a) 
  else
  begin
    res:=part(part(allinfinitesimals,1),length(xvar)+a) $
    der:=part(temp_uvar,a) $
    for k1:=1:length(lis) do
    begin
      temp:=for k2:=1:length(xvar) sum
        dlie(part(part(allinfinitesimals,1),k2),part(xvar,part(lis,k1)))*
             df(der,part(xvar,k2)) $
      der:=df(der,part(xvar,part(lis,k1))) $
      res:=dlie(res,part(xvar,part(lis,k1)))-temp $
    end $
  end $  
  return arrange(res) $
end $
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
procedure mux(a,lis) $
% Prolonged infinitesimal generator of arbitrary elements
begin
  scalar xietamu,der,res,temp,vars,tempind $
  integer k1,k2 $
  vars:=part(zvar,2) $
  xietamu:={} $
  for k1:=1:length(vars) do
  begin
    tempind:=part(part(zvar,1),k1) $
    xietamu:=append(xietamu,{etax(part(tempind,1),part(tempind,2))}) $
  end $
  res:=part(part(allinfinitesimals,1),length(xvar)+length(uvar)+a) $
  if lis neq {0} then
  begin
    der:=part(temp_arbelem,a) $
    for k1:=1:length(lis) do
    begin
      temp:=for k2:=1:length(vars) sum
        dliestar(part(xietamu,k2),part(vars,part(lis,k1)))*df(der,part(vars,k2)) $
      der:=df(der,part(vars,part(lis,k1))) $
      res:= dliestar(res,part(vars,part(lis,k1)))-temp $
    end $
  end $  
  return arrange(res) $
end $
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
procedure etax_approx(a,lis) $
% Prolonged approximate infinitesimal generator
begin
  scalar der,res,temp $
  integer k1,k2 $
  if lis={-1} then
    res:=part(part(allinfinitesimals,1),a)
  else if lis={0} then
    res:=part(part(allinfinitesimals,1),length(xvar)+a) 
  else
  begin
    res:=part(part(allinfinitesimals,1),length(xvar)+a) $
    der:=part(temp_uvar,a) $
    for k1:=1:length(lis) do
    begin
      temp:=for k2:=1:length(xvar) sum
        dlieapprox(part(part(allinfinitesimals,1),k2),part(xvar,part(lis,k1)))*
                   df(der,part(xvar,k2)) $
      der:=df(der,part(xvar,part(lis,k1))) $
      res:= dlieapprox(res,part(xvar,part(lis,k1)))-temp $
    end $
  end $  
  res:=substitute_expansion_uvar(res) $ 
  return arrangeapprox(res) $
end $
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
procedure relieinv $
% Computes the infinitesimal invariance condition
begin
  if (steprelie geq 1) then
  begin 
    if membership(part(okinputrelie,6),{3,7})=1 then  
    % diffeqs and leadders or lagrangian are ok
    begin  
      if part(okinputrelie,5)=0 then  % exact symmetries
        relieinv_exact()
      else
        relieinv_approx() $ 
      steprelie:=2 $  
      write "'invcond' has been computed (invariance conditions)." $
    end 
    else
      write "Check diffeqs and/or leadders, or the Lagrangian!" $    
  end 
  else
    write "relieinit() has not been called yet!" $    
  end $
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
procedure relieinv_exact $
begin
  scalar deq_obj,inv_obj,der_obj,tempder,tempeqn,combinations,
         subst,neweqn,newder,solut,lis,kk $
  integer k1,k2,k3,k4 $
  invcond:={} $
  if  part(okinputrelie,6)=7 then % variational symmetries
  begin
    deq_obj:=part(lagrangian,1) $
    inv_obj:=0 $
    for k1:=1:length(part(jet,1)) do
    begin
      der_obj:=part(part(jet,3),k1) $
      tempder:=df(deq_obj,der_obj) $
      if tempder neq 0 then
      begin
        kk:=part(part(part(jet,1),k1),1) $
        lis:=part(part(part(jet,1),k1),2) $
        inv_obj:=inv_obj+etax(kk,lis)*tempder  
      end $
    end $
    for k1:=1:length(xvar) do
      inv_obj:=inv_obj+dlie(part(part(allinfinitesimals,1),k1),part(xvar,k1))*deq_obj
               +dlie(part(part(allphi,1),k1),part(xvar,k1)) $
    inv_obj:=arrange(inv_obj) $  
    inv_obj:=num(inv_obj) $
    invcond:=append(invcond,{inv_obj}) $
  end
  else
  begin
    for k1:=1:length(diffeqs) do
    begin
      deq_obj:=part(diffeqs,k1) $
      inv_obj:=0 $
      for k2:=1:length(part(jet,1)) do
      begin
        der_obj:=part(part(jet,3),k2) $
        tempder:=df(deq_obj,der_obj) $
        if tempder neq 0 then
        begin
          kk:=part(part(part(jet,1),k2),1) $
          lis:=part(part(part(jet,1),k2),2) $
          inv_obj:=inv_obj+etax(kk,lis)*tempder $
        end $
      end $
      if part(okinputrelie,2)=1 then % equivalence transformations
      begin
        for k2:=1:length(part(jet_equiv,1)) do
        begin
          der_obj:=part(part(jet_equiv,3),k2) $
          temp_der:=df(deq_obj,der_obj) $
          if temp_der neq 0 then
          begin
            kk:=part(part(part(jet_equiv,1),k2),1) $
            lis:=part(part(part(jet_equiv,1),k2),2) $
            inv_obj:=inv_obj+mux(kk,lis)*temp_der $ 
          end $
        end $
      end $
      invcond:= append(invcond,{inv_obj}) $
    end $
    invcond:=sub(solutiondedv,invcond) $
    invcond:=for k1:=1:length(invcond) collect num(part(invcond,k1)) $
  end $ 
end $
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
procedure relieinv_approx $
begin
  scalar deq_obj,inv_obj,der_obj,tempder,tempeqn,allde,alldv,subst,
         neweqn,newder,combinations,newdiffeqs,newleadders,kk,lis,
         temp1,temp2 $
  integer k1,k2,k3,k4 $
  if  part(okinputrelie,6)=7 then % variational symmetries
  begin
    deq_obj:=part(lagrangian,1) $
    inv_obj:=0 $
    for k2:=1:length(part(jetsplit,1)) do
    begin
      der_obj:=part(part(jetsplit,2),k2) $
      tempder:=df(deq_obj,der_obj) $
      if tempder neq 0 then
      begin
        kk:=part(part(part(jetsplit,1),k2),1) $
        lis:=part(part(part(jetsplit,1),k2),2) $
        inv_obj:=inv_obj+etax_approx(kk,lis)*tempder 
      end $
    end $
    for k1:=1:length(xvar) do
      inv_obj:=inv_obj
        +dlieapprox(part(part(allinfinitesimals,1),k1),part(xvar,k1))*deq_obj
        +dlieapprox(part(part(allphi,1),k1),part(xvar,k1)) $
    inv_obj:=substitute_expansion_uvar(inv_obj) $    
    inv_obj:=arrangeapprox(inv_obj) $  
    invcond:={num(inv_obj)} $
  end
  else
  begin  
    invcond:={} $
    diffeqs:=inversearrange(diffeqs) $
    diffeqs:=substitute_expansion_uvar(diffeqs) $
    diffeqs:=arrangeapprox(diffeqs) $
    leadders:=inversearrange(leadders) $
    leadders:=substitute_expansion_uvar(leadders) $
    leadders:=arrangeapprox(leadders) $
    nonpolyders:=inversearrange(nonpolyders) $
    nonpolyders:=substitute_expansion_uvar(nonpolyders) $
    nonpolyders:=arrangeapprox(nonpolyders) $
    for k1:=1:length(diffeqs) do
    begin
      deq_obj:=part(diffeqs,k1) $
      inv_obj:=0 $
      for k2:=1:length(part(jetsplit,1)) do
      begin
        der_obj:=part(part(jetsplit,2),k2) $
        tempder:=df(deq_obj,der_obj) $
        if tempder neq 0 then
        begin
          kk:=part(part(part(jetsplit,1),k2),1) $
          lis:=part(part(part(jetsplit,1),k2),2) $
          inv_obj:=inv_obj+etax_approx(kk,lis)*tempder $
        end $
      end $
      invcond:= append(invcond,{inv_obj}) $
    end $
    invcond:=sub(solutiondedv,invcond) $
    invcond:=for k1:=1:length(diffeqs) collect num(part(invcond,1)) $
  end $  
end $
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
procedure reliedet $
% Computes the determining equations for infinitesimals
begin
  if (steprelie geq 2) then
  begin
    if part(okinputrelie,5)=0 then % exact symmetries
      reliedet_exact()
    else
      reliedet_approx() $  
    steprelie:=3 $  
    write "'deteqs' have been computed (determining equations)." $
  end
  else       
    write "relieinv() has not been called yet!" $    
end $
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
procedure reliedet_exact $
begin
  scalar vars,newjet,newjet_equiv $
  integer k1,nvar $
  if (part(okinputrelie,6) = 7) then    % variational symmetries
    nvar:=length(xvar)+length(uvar)*bincoeff(length(xvar)+jetorder-1,jetorder-1)
  else
  begin
    if (part(okinputrelie,4) = 1) then  % contact symmetries
      nvar:=2*length(xvar)+1
    else                                  % point symmetries
      nvar:=length(xvar)+length(uvar) $  
  end $    
  vars:=for k1:=1:nvar collect part(part(jet,3),k1) $
  vars:=append(vars,nonpolyders) $
  for k1:=1:length(freepars) do
    vars:=append(vars,fargs(part(freepars,k1))) $
  newjet:=part(jet,3) $ 
  newjet:=removemultiple(newjet,vars) $
  if part(okinputrelie,2) = 1 then  % equivalence transformations
  begin
    newjet_equiv:=part(jet_equiv,3) $
    newjet_equiv:=removemultiple(newjet_equiv,arbelem) $
    newjet:=append(newjet,newjet_equiv) $
  end $
  deteqs:=allcoeffs(invcond,newjet) $     
  deteqs:=delzero(deteqs) $
end $
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
procedure reliedet_approx $
begin
  scalar vars,newjet,temp_nonpolyders $
  integer k1,njet1,njetm1 $
  if part(okinputrelie,6) = 7 then  % variational symmetries
  begin
    njetm1:=length(xvar)+length(uvar)*bincoeff(length(xvar)+jetorder-1,jetorder-1) $
    vars:=for k1:=1:njetm1 collect part(part(jetapprox,2),k1) $
  end 
  else 
  begin
    if part(okinputrelie,4) = 1 then  % contact symmetries
    begin
      njet1:=2*length(xvar)+1 $
      vars:=for k1:=1:njet1 collect part(part(jetapprox,2),k1) $  
    end 
    else
      vars:=append(xvar,part(uvarall,1)) $
  end $    
  for k1:=1:length(freepars) do
    vars:=append(vars,fargs(part(freepars,k1))) $
  newjet:=part(jetapprox,2) $
  newjet:=removemultiple(newjet,vars) $
  temp_nonpolyders:=allcoeffs(nonpolyders,{epsilon}) $
  newjet:=removemultiple(newjet,temp_nonpolyders) $
  newjet:=append(newjet,{epsilon}) $ 
  deteqs:=allcoeffs(invcond,newjet) $ 
  deteqs:=delzero(deteqs) $  
end $
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
procedure reliesolve $
% Solves the determining equations using CRACK package
begin
  if (steprelie geq 3) then
  begin
    if part(okinputrelie,5)=0 then  % exact symmetries
      reliesolve_exact()
    else
      reliesolve_approx() $           % approximate symmetries
    write "'symmetries' have been computed (infinitesimals)." $
    steprelie:=4 $
  end
  else       
    write "reliedet() has not been called yet!" $    
end $
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
procedure reliesolve_exact $
begin
  scalar allunks,solsym,pars,par,temppar,finitepar,infinitepar,subst1,subst2,subst,
         auxvars,elimvars $
  integer k1,k2 $
  allunks:= foreach x in part(allinfinitesimals,2) join
      if membership(x,{0,1})>0 then {} else {x} $
  allunks:=append(allunks,freepars) $
  allunks:=append(allunks,part(allphi,2)) $
  elimvars:={} $
  for k1:=1:length(allunks) do
    elimvars:=append(elimvars,fargs(part(allunks,k1))) $
  auxvars:={} $
  for k1:=1:length(nonpolyders) do
    if membership(part(nonpolyders,k1),elimvars)=0 then
      auxvars:=append(auxvars,{part(nonpolyders,k1)}) $
  solsym:=crack(deteqs,nonzeropars,allunks,auxvars) $
  pars:={} $
  for k1:=1:length(solsym) do
  begin  
    temppar:=part(part(solsym,k1),3) $ % parameters involved in the infinitesimals
    for k2:=1:length(temppar) do
    begin
      if (membership(part(temppar,k2),append(pars,allunks))=0) then 
        pars:=append(pars,{part(temppar,k2)}) $
    end $
  end $
  pars:=removemultiple(pars,freepars) $
  finitepar:={} $
  infinitepar:={} $
  foreach par in pars do
    if (fargs(par) = {}) then 
      finitepar:=append(finitepar,{par})
    else
      infinitepar:=append(infinitepar,{par}) $
  arbconst:=letterlist(k_,length(finitepar)) $
  arbfun:=letterlist(f_,length(infinitepar)) $
  for k1:=1:length(infinitepar) do
    dependence({part(arbfun,k1)},fargs(part(infinitepar,k1))) $
  subst1:=for k1:=1:length(finitepar) collect
    part(finitepar,k1)=part(arbconst,k1) $  
  subst2:=for k1:=1:length(infinitepar) collect
    part(infinitepar,k1)=part(arbfun,k1) $
  subst:=append(subst1,subst2) $ 
  symmetries:=sub(subst,solsym) $
end $
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
procedure reliesolve_approx $
begin
  scalar allunks,allpars,solsym,pars,par,temppar,finitepar,infinitepar,
         subst1,subst2,subst,auxvars,elimvars $
  integer k1,k2 $
  allunks:=foreach x in part(allinfinitesimals,2) join
    if membership(x,{0,1})>0 then {} else {x} $
  allunks:=append(allunks,freepars) $
  allunks:=append(allunks,part(allphi,2)) $
  allpars:=allcoeffs(nonpolyders,{epsilon}) $
  elimvars:={} $
  for k1:=1:length(allunks) do
    elimvars:=append(elimvars,fargs(part(allunks,k1))) $
  auxvars:={} $
  %for k1:=1:length(allpars) do
  %  if membership(part(allpars,k1),elimvars)=0 then
  %    auxvars:=append(auxvars,{part(allpars,k1)}) $
  solsym:=crack(deteqs,nonzeropars,allunks,auxvars) $
  pars:={} $
  for k1:=1:length(solsym) do
  begin  
    temppar:=part(part(solsym,k1),3) $ % parameters involved in the infinitesimals
    for k2:=1:length(temppar) do
    begin
      if (membership(part(temppar,k2),append(pars,allunks))=0) then 
        pars:=append(pars,{part(temppar,k2)}) $
    end $
  end $
  pars:=removemultiple(pars,freepars) $
  finitepar:={} $
  infinitepar:={} $
  foreach par in pars do
    if (fargs(par) = {}) then 
      finitepar:=append(finitepar,{par})
    else
      infinitepar:=append(infinitepar,{par}) $  
  arbconst:=letterlist(k_,length(finitepar)) $
  arbfun:=letterlist(f_,length(infinitepar)) $
  for k1:=1:length(infinitepar) do
    dependence({part(arbfun,k1)},fargs(part(infinitepar,k1))) $
  subst1:=for k1:=1:length(finitepar) collect
    part(finitepar,k1)=part(arbconst,k1) $  
  subst2:=for k1:=1:length(infinitepar) collect
    part(infinitepar,k1)=part(arbfun,k1) $
  subst:=append(subst1,subst2) $ 
  symmetries:=sub(subst,solsym) $
end $
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
procedure reliegen(k,lis) $
% Splits the symmetries finding a basis of Lie algebra of symmetries
begin
  if (steprelie geq 4) then
  begin
    if (k>0 and k leq length(symmetries)) then
    begin
      if part(okinputrelie,5)=0 then    % Exact symmetries
        reliegen_exact(k,lis)      
      else                                % Approximate symmetries  
        reliegen_approx(k,lis) $  
      write "'generators' have been computed (vector fields of symmetries)." $
      if part(okinputrelie,6)=7 then
      begin
        write "'cogenerators' have been computed (functions for variatiational symmetries)." $
        steprelie:=5 $
      end $
    end
    else
      write "The requested symmetry does not exist!" $  
  end    
  else
    write "reliesolve() has not been called yet!" $
end $
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
procedure reliegen_exact(k,lis) $
begin
  scalar solution,tempgen,tempphi,temppar,lis1,subst,subst1,subst2 $
  integer k1,k2 $
  generators:={} $
  cogenerators:={} $
  solution:=part(symmetries,k) $
  if lis = {-1} then
  begin
    generators:={sub(part(solution,2),part(allinfinitesimals,1))} $
    cogenerators:={sub(part(solution,2),part(allphi,1))} $ 
  end 
  else
  begin
    tempgen:=sub(part(solution,2),part(allinfinitesimals,1)) $
    tempphi:=sub(part(solution,2),part(allphi,1)) $
    temppar:=part(solution,3) $
    temppar:=removemultiple(temppar,freepars) $
    temppar:=removemultiple(temppar,arbfun) $
    temppar:=removemultiple(temppar,part(solution,4)) $
    if length(lis) neq length(temppar) then
      lis1:=for k1:=1:length(temppar) collect 1
    else
      lis1:=lis $  
    for k1:=1:length(temppar) do
    begin
      subst1:=for k2:=1:length(temppar) collect
        part(temppar,k2)=kroneckerdelta(k1,k2)*part(lis1,k1) $
      subst2:=for k2:=1:length(arbfun) collect
        part(arbfun,k2)=0 $
      subst:=append(subst1,subst2) $
      if delzero(sub(subst,tempgen)) neq {} then
      begin
        generators:=append(generators,{sub(subst,tempgen)}) $
        cogenerators:=append(cogenerators,{sub(subst,tempphi)}) $    
      end $
    end $
  end $
 end $ 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
procedure reliegen_approx(k,lis) $
begin
  scalar solution,tempgen,tempphi,temppar,lis1,subst,subst1,subst2 $
  integer k1,k2 $
  generators:={} $
  if (k > 0) then
  begin
    solution:=part(symmetries,k) $
    if lis = {-1} then
    begin
      generators:={sub(part(solution,2),part(allinfinitesimals,1))} $
      cogenerators:={sub(part(solution,2),part(allphi,1))} $ 
    end 
    else
    begin
      tempgen:=sub(part(solution,2),part(allinfinitesimals,1)) $
      tempphi:=sub(part(solution,2),part(allphi,1)) $
      temppar:=part(solution,3) $
      temppar:=removemultiple(temppar,freepars) $
      temppar:=removemultiple(temppar,arbfun) $
      temppar:=removemultiple(temppar,part(solution,4)) $
      generators:={} $
      cogenerators:={} $
      if length(lis) neq length(temppar) then
        lis1:=for k1:=1:length(temppar) collect 1
      else
        lis1:=lis $   
      for k1:=1:length(temppar) do
      begin
        subst1:=for k2:=1:length(temppar) collect
                part(temppar,k2)=kroneckerdelta(k1,k2)*part(lis1,k1) $
        subst2:=for k2:=1:length(arbfun) collect
                part(arbfun,k2)=0 $
        subst:=append(subst1,subst2) $
        if delzero(sub(subst,tempgen)) neq {} then
        begin
          generators:=append(generators,{sub(subst,tempgen)}) $
          cogenerators:=append(cogenerators,{sub(subst,tempphi)}) $    
        end $ 
      end $     
    end $
  end $
end $
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
procedure liebracket(obj1,obj2) $
% Computes the Lie bracket of two generators
begin
  scalar temp,res,vars $
  integer k1,k2 $
  if steprelie > 0 then
  begin
    if part(okinputrelie,5)=0 then      % exact symmetries
    begin
      vars:=append(xvar,uvar) $
      if part(okinputrelie,2) = 1 then  % equivalence transformations
        vars:=append(vars,arbelem) $
      if part(okinputrelie,4) = 1 then  % exact contact symmetries
        vars:=for k1:=1:2*length(xvar)+1 collect part(part(jet,3),k1) $
    end  
    else                                  % approximate symmetries
    begin
      vars:=append(xvar,part(uvarall,1)) $
      if part(okinputrelie,4) = 1 then  % approximate contact symmetries
        vars:=for k1:=1:2*length(xvar)+1 collect part(part(jetapprox,2),k1) $
    end $  
    res:={} $
    for k1:=1:length(vars)  do
    begin
      temp:=for k2:=1:length(vars) sum
        (part(obj1,k2)*df(part(obj2,k1),part(vars,k2))
           - part(obj2,k2)*df(part(obj1,k1),part(vars,k2))) $
      res:=append(res,{temp}) $  
    end $ 
  end  
  else
  begin
    write "relieinit() has not been called yet." $
    res:={} $
  end $  
  return res $
end $
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
procedure structureconstants(vectorfields,i,j) $
% Returns the components with respect to the basis given by vectorfields
% of the Lie bracket between k1-th and k2-th element of vectorfields 
begin
  scalar constr $
  if steprelie > 0 then
  begin
    if (part(okinputrelie,5) = 0) then       % exact symmetries
      constr:=structureconstants_exact(vectorfields,i,j) 
    else                                       % approximate symmetries
      constr:=structureconstants_approx(vectorfields,i,j) $   
  end
  else
  begin
    write "relieinit() has not been called yet." $
    constr:={} $
  end $ 
  return constr $
end $
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
procedure structureconstants_exact(vectorfields,i,j) $
begin
  scalar vars,constr,lbz,solut,cffs $
  integer k1,k2,k3 $
  vars:=append(xvar,uvar) $
  cffs:=letterlist(cff,length(vectorfields));
  if part(okinputrelie,2) = 1 then 
    vars:=append(vars,arbelem) $
  if part(okinputrelie,4) = 1 then 
    vars:=for k1:=1:2*length(xvar)+1 collect part(part(jet,3),k1) $
  lbz:=liebracket(part(vectorfields,j),part(vectorfields,i)) $ 
  for k1:=1:length(vectorfields) do
    lbz:=sumlist(lbz,scalarmult(part(cffs,k1),part(vectorfields,k1))) $
  lbz:=delzero(lbz) $
  if length(lbz)=0 then
    constr:=zerolist(length(vectorfields)) 
  else
  begin
    symbolic <<;print_:=nil;>> $
    solut:=part(part(crack(lbz,{},cffs,vars),1),2) $
    constr:=sub(solut,cffs) $
    symbolic <<;print_:=1;>> $
  end $  
  return constr $
end $
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
procedure structureconstants_approx(vectorfields,i,j) $
% Computes the list of structure constants for the Lie bracket of two generators
begin
  scalar vars,constr,z,z1,z2,z3,zz,solut,cffs $
  integer k1,k2,k3 $
  cffs:=letterlist(cff,length(vectorfields));
  vars:=append(xvar,part(uvarall,1)) $
  if part(okinputrelie,4) = 1 then  % approximate contact symmetries
        vars:=for k1:=1:2*length(xvar)+1 collect part(part(jetapprox,2),k1) $
  vars:=append(vars,{epsilon}) $
  lbz:=liebracket(part(vectorfields,j),part(vectorfields,i)) $ 
  for k1:=1:length(vectorfields) do
    lbz:=sumlist(lbz,scalarmult(part(cffs,k1),part(vectorfields,k1))) $
  if length(lbz)=0 then
    constr:=zerolist(length(vectorfields)) 
  else
  begin
    solut:=part(part(crack(lbz,{},cffs,vars),1),2) $
    constr:=sub(solut,cffs) $
  end $
  return constr $  
end $
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
procedure commutatortable(vectorfields) $
% Computes the commutator table of vectorfields
begin
  scalar c,cz,temp $
  integer k1,k2,nvf $
  nvf:=length(vectorfields) $
  cz:=letterlist(vf_,nvf) $
  commtable:={} $
  for k1:=1:nvf do
  begin  
    temp:={} $
    for k2:=1:nvf do
    begin
      if k1=k2 then
        temp:=append(temp,{0})
      else if k1>k2 then
        temp:=append(temp,{-part(part(commtable,k2),k1)})
      else     
      begin 
        c:=structureconstants(vectorfields,k1,k2) $
        temp:=append(temp,{scalarproduct(c,cz)}) $
      end $  
    end $
    commtable:=append(commtable,{temp}) $
  end $
  write "'commtable' has been computed (commutator table)." $
  return commtable $
end $
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
procedure nonzerocommutators(vectorfields) $
% Prints the non-zero Lie brackets of vectorfields
begin
  scalar cz,ct,temp $
  integer k1,k2,nvf $
  nvf:=length(vectorfields) $
  cz:=letterlist(vf_,nvf) $
  nzcomm:={} $
  ct:=commutatortable(vectorfields) $
  for k1:=1:nvf-1 do
  begin
    for k2:=k1+1:nvf do
    begin
      temp:={{part(cz,k1),part(cz,k2)},part(part(ct,k1),k2)} $ 
      if part(part(ct,k1),k2) neq 0 then
      begin
        nzcomm:=append(nzcomm,{temp}) $
      end $
    end $  
  end $
  write "'nzcomm' has been computed (list of non-zero commutators)." $
  return nzcomm $
end $        
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
procedure relieprol $
% Computes the prolongation of a generic infinitesimal generator
begin
  scalar lis,newjet $
  integer k1,a $
  if (steprelie geq 1) then
  begin
    newjet:=part(jet,3) $
    prolongation:={} $
    for k1:=1:length(part(jet,1)) do
    begin
      a:=part(part(part(jet,1),k1),1) $
      lis:=part(part(part(jet,1),k1),2) $
      if part(okinputrelie,5)=0 then
        prolongation:=append(prolongation,{etax(a,lis)})  
      else
        prolongation:=append(prolongation,{etax_approx(a,lis)}) $   
    end $ 
    if part(okinputrelie,2)=1 then
    begin
      for k1:=1:length(part(jet_equiv,1)) do
      begin
        a:=part(part(part(jet_equiv,1),k1),1) $
        lis:=part(part(part(jet_equiv,1),k1),2) $
        prolongation:=append(prolongation,{mux(a,lis)}) $
      end $
      newjet:=append(newjet,part(jet_equiv,3)) $  
    end $ 
    prolongation:={newjet,prolongation} $
    write "'prolongation' has been computed." $   
  end
  else
  begin
    write "relieinit() has not been called yet!" $
    prolongation:={} $
  end $  
  return prolongation $  
end $
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
procedure reliedistrib(vectorfields) $
% computes the distribution of a list of generators until the order jetorder
begin
  scalar temp,xieta,substitution $
  integer k1,k2,nr,nc $
  if (steprelie geq 1) then
  begin
    relieprol() $
    temp:=part(prolongation,2) $
    nc:=length(temp);
    nr:=length(vectorfields) $
    matrix distribution(nr,nc) $
    xieta:=part(allinfinitesimals,1) $
    for k1:=1:nr do
    begin
      substitution:={} $
      for k2:=1:length(xieta) do
        substitution:=append(substitution,
                      {part(xieta,k2)=part(part(vectorfields,k1),k2)}) $
      for k2:=1:nc do
        distribution(k1,k2):=sub(substitution,part(temp,k2)) $
    end $
  end
  else
    write "relieinit() has not been called yet!" $
  write "the matrix 'distribution' has been computed." $
  return distribution $
end $
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
procedure relie(flag) $
% General function for Lie symmetry analysis of differential equations
begin
  if (flag > 0) then 
  begin
    write "Calling relieinit()..." $
    relieinit() $
  end $  
  if (flag > 1) then 
  begin
    write "Calling relieinv()..." $
    relieinv() $
  end $  
  if (flag > 2) then 
  begin
    write "Calling reliedet()..." $
    reliedet() $
  end $  
  if (flag > 3) then   
  begin
    write "Calling reliesolve()..." $
    reliesolve() $
  end $  
  if (flag > 4) then   
    write "Invalid argument for relie()" $
end $
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
procedure testrank(gens) $
begin
  integer k1 $
  scalar jetvar,rndlst,subst,temp $
  if (steprelie geq 1) then
  begin
    reliedistrib(gens) $
    jetvar:=part(jet,3) $
    rndlst:=for k1:=1:length(jetvar) collect random(2*length(jetvar))+1 $
    subst:=for k1:=1:length(jetvar) collect part(jetvar,k1)=part(rndlst,k1) $
    temp:=sub(subst,distribution) $
    rankdistr:=rank(temp) $
    write "'rankdistr' has been computed." $
  end  
  else
    write "'relieinit() has not been called yet!" $
end $
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
procedure generatealgebra(k) $
begin
  integer k1,k2,k3 $
  scalar vars,cmbs,cmb,temp $
  if (steprelie geq 1) then
  begin
    vars:=append(xvar,uvar) $
    generators:={} $ 
    if k = 1 then         % isometries
    begin
      cmbs:=combrep(length(vars),2) $
      for k1:=1:length(vars) do
      begin  
        temp:=for k2:=1:length(vars) collect kroneckerdelta(k1,k2) $
        generators:=append(generators,{temp}) $ 
      end $
      for k1:=1:length(cmbs) do
      begin
        cmb:=part(cmbs,k1) $
        if part(cmb,1) neq part(cmb,2) then
        begin
          temp:={} $
          for k2:=1:length(vars) do
          begin
            if part(cmb,1) = k2 then
              temp:=append(temp,{part(vars,part(cmb,2))}) 
            else if part(cmb,2) = k2 then
              temp:=append(temp,{-part(vars,part(cmb,1))}) 
            else
              temp:=append(temp,{0}) $  
          end $
          generators:=append(generators,{temp}) $
        end $
      end $
    end
    else if k = 2 then    % affine transformations
    begin
      for k1:=1:length(vars) do
      begin  
        temp:=for k2:=1:length(vars) collect kroneckerdelta(k1,k2) $
        generators:=append(generators,{temp}) $ 
      end $
      for k1:=0:length(vars)-1 do
      begin
        for k2:=1:length(vars) do
        begin  
          temp:=zerolist(k1);
          temp:=append(temp,{part(vars,k2)}) $
          for k3:=1:length(vars)-k1-1 do
            temp:=append(temp,{0}) $
          generators:=append(generators,{temp}) $
        end $
      end $
    end
    else if obj = k then    % projective transformations
    begin
      for k1:=1:length(vars) do
      begin  
        temp:=for k2:=1:length(vars) collect kroneckerdelta(k1,k2) $
        generators:=append(generators,{temp}) $ 
      end $
      for k1:=0:length(vars)-1 do
      begin
        for k2:=1:length(vars) do
        begin  
          temp:=zerolist(k1);
          temp:=append(temp,{part(vars,k2)}) $
          for k3:=1:length(vars)-k1-1 do
            temp:=append(temp,{0}) $
          generators:=append(generators,{temp}) $
        end $
      end $
      for k1:=1:length(vars) do
        generators:=append(generators,{scalarmult(part(vars,k1),vars)}) $
    end $
    write "The list 'generators' has been computed." $
    write generators $
  end 
  else
    write "'relieinit() has not been called yet!" $
end $  
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
procedure abelian(gens) $
% tests if a set of generators spans an Abelian Lie algebra
begin
  scalar temp $
  temp:=nonzerocommutators(gens) $
  if length(temp)=0 then
    write "The Lie Algebra is Abelian." 
  else
    write "The Lie Algebra is NOT Abelian." $
end $
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
procedure solvable(gens) $
% tests if a set of generators spans a solvable Lie algebra
begin
  scalar temp,newgens,lb,temp $
  integer niter,kk,k1,k2 $
  newgens:=gens $
  niter:=length(newgens) $
  kk:=0 $
  while ((length(newgens) neq 0) and (kk < niter)) do
  begin
    temp:={} $
    for k1:=1:length(newgens)-1 do
      for k2:=k1+1:length(newgens) do
      begin
        lb:=liebracket(part(newgens,k1),part(newgens,k2)) $
        if length(delzero(lb)) neq 0 then
          temp:=append(temp,{lb}) $
      end $
    newgens:=temp $
    kk:=kk+1 $
  end $
  if (length(newgens) = 0) then
    write "The Lie Algebra is Solvable." 
  else
    write "The Lie Algebra is NOT Solvable." $
end $
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
procedure etaevolutionary_exact(gen,a) $
begin
  scalar temp $
  integer k1 $
  temp:=for k1:=1:length(xvar) sum part(gen,k1)*df(part(temp_uvar,a),part(xvar,k1)) $
  temp:=part(gen,length(xvar)+a)-temp $
  return arrange(temp) $
end $  
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
procedure etaevolutionary_approx(gen,a) $
begin
  scalar temp $
  integer k1 $
  temp:=for k1:=1:length(xvar) sum 
    part(gen,k1)*df(part(temp_uvar,a),part(xvar,k1)) $
  temp:=part(gen,length(xvar)+a)-temp $
  temp:=substitute_expansion_uvar(temp) $
  return arrangeapprox(temp) $
end $  
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
procedure derlagrangian(obj,a,lis) $
begin
  scalar var,dervar $
  integer k1 $
  dervar:=part(temp_uvar,a) $
  for k1:=1:length(lis) do
  begin
    var:=part(xvar,part(lis,k1)) $
    dervar:=df(dervar,var) $
  end $  
  return df(obj,arrange(dervar)) $
end $
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
procedure dlien(obj,lis) $
begin
  integer k1 $
  scalar temp $
  temp:=obj $
  for k1:=1:length(lis) do
    temp:=dlie(temp,part(xvar,part(lis,k1))) $
  return temp $    
end $
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
procedure dlien_approx(obj,lis) $
begin
  integer k1 $
  scalar temp $
  temp:=obj $
  for k1:=1:length(lis) do
    temp:=dlieapprox(temp,part(xvar,part(lis,k1))) $
  return temp $    
end $
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
procedure dervar(a,lis) $
begin
  scalar dep,var $
  integer k1 $
  dep:=part(temp_uvar,a) $
  for k1:=1:length(lis) do
  begin
    var:=part(xvar,part(lis,k1)) $
    dep:=df(dep,var) $
  end $
  return arrange(dep) $
end $  
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
procedure dervar_approx(a,lis) $
begin
  scalar dep,var $
  integer k1 $
  dep:=part(part(temp_uvarall,1),a) $
  for k1:=1:length(lis) do
  begin
    var:=part(xvar,part(lis,k1)) $
    dep:=df(dep,var) $
  end $
  return arrangeapprox(dep) $
end $  
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
procedure compute_intermediate_flux_exact(lagr,gen,a,i) $
begin
  scalar temp,temp1,temp2,comb,combi,combb,etaevol,etaevoltemp,res $
  integer k1,k2,k3,k4,sign $
  res:=0 $
  etaevol:={etaevolutionary_exact(gen,a)} $
  for k1:=0:jetorder-1 do
  begin
    sign:=1 $
    comb:=combrep(length(xvar),k1) $
    combi:=for k2:=1:length(comb) collect append({i},part(comb,k2)) $
    temp:={} $
    for k2:=1:length(combi) do
    begin
      dep:=dervar(a,part(combi,k2)) $
      temp:=append(temp,{df(lagr,dep)}) $
    end $
    res:=res+sign*scalarproduct(etaevol,temp) $
    for k2:=1:jetorder-k1-1 do
    begin
      temp:={} $
      comb:=combrep(length(xvar),k2) $
      for k3:=1:length(combi) do
      begin
        sign:=-sign $
        combb:=for k4:=1:length(comb) collect 
          append(part(combi,k3),part(comb,k4)) $
        temp1:={} $
        for k4:=1:length(combb) do
        begin
          dep:=dervar(a,part(combb,k4)) $
          temp1:=append(temp1,{dlien(df(lagr,dep),part(comb,k4))}) $
        end $
        temp2:=for k4:=1:length(temp1) sum part(temp1,k4) $
        temp:=append(temp,{temp2}) $
      end $  
      res:=res+sign*scalarproduct(etaevol,temp) $
    end $  
    etaevoltemp:={} $
    for k2:=1:length(etaevol) do
      for k3:=1:length(xvar) do 
        etaevoltemp:=append(etaevoltemp,{dlie(part(etaevol,k2),part(xvar,k3))}) $
    etaevol:=etaevoltemp $
  end $
  return res $
end $
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
procedure compute_flux_exact(lagr,gen,cogen) $
begin
  integer k1,k2 $
  scalar temp,temp1,res $
  res:={} $
  for k1:=1:length(xvar) do
  begin
    temp:=for k2:=1:length(uvar) sum 
      compute_intermediate_flux_exact(lagr,gen,k2,k1) $
    temp:=temp+part(cogen,k1) $
    temp1:=for k2:=1:length(xvar) collect part(gen,k2)*lagr $
    temp:=temp+part(temp1,k1) $
    res:=append(res,{temp}) $
  end $
  return res $  
end $
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
procedure compute_intermediate_flux_approx(lagr,gen,a,i) $
begin
  scalar temp,temp1,temp2,comb,combi,combb,etaevol,etaevoltemp,res $
  integer k1,k2,k3,k4,sign $
  res:=0 $
  etaevol:={etaevolutionary_approx(gen,a)} $
  for k1:=0:jetorder-1 do
  begin
    sign:=1 $
    comb:=combrep(length(xvar),k1) $
    combi:=for k2:=1:length(comb) collect append({i},part(comb,k2)) $
    temp:={} $
    for k2:=1:length(combi) do
    begin
      dep:=dervar_approx(a,part(combi,k2)) $
      temp:=append(temp,{df(lagr,dep)}) $
    end $
    res:=res+sign*scalarproduct(etaevol,temp) $
    for k2:=1:jetorder-k1-2 do
    begin
      temp:={} $
      comb:=combrep(length(xvar),k2) $
      for k3:=1:length(combi) do
      begin
        sign:=-sign $
        combb:=for k4:=1:length(comb) collect 
          append(part(combi,k3),part(comb,k4)) $
        temp1:={} $
        for k4:=1:length(combb) do
        begin
          dep:=dervar_approx(a,part(combb,k4)) $
          temp1:=append(temp1,{dlien_approx(df(lagr,dep),part(comb,k4))}) $
        end $
        temp2:=for k4:=1:length(temp1) sum part(temp1,k4) $
        temp:=append(temp,{temp2}) $
      end $  
      res:=res+sign*scalarproduct(etaevol,temp) $
    end $  
    etaevoltemp:={} $
    for k2:=1:length(etaevol) do
      for k3:=1:length(xvar) do 
        etaevoltemp:=append(etaevoltemp,{
          dlieapprox(part(etaevol,k2),part(xvar,k3))}) $
    etaevol:=etaevoltemp $
  end $
  return res $
end $
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
procedure compute_flux_approx(lagr,gen,cogen) $
begin
  integer k1,k2 $
  scalar temp,res $
  res:={} $
  for k1:=1:length(xvar) do
  begin
    temp:=for k2:=1:length(uvar) sum 
      compute_intermediate_flux_approx(lagr,gen,k2,k1) $
    temp:=temp+part(cogen,k1) $
    temp1:=for k2:=1:length(xvar) collect part(gen,k2)*lagr $
    temp:=temp+part(temp1,k1) $
    res:=append(res,{temp}) $
  end $
  return res $  
end $
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
procedure relieclaw(kk) $
begin
  if (steprelie >= 5) then
  begin
    if part(okinputrelie,6) = 7 then
    begin
      if part(okinputrelie,5)=0 then   % exact variational symmetries
        fluxes:=compute_flux_exact(part(lagrangian,1),part(generators,kk),part(cogenerators,kk)) 
      else                               % approximate variational symmetries
        fluxes:=compute_flux_approx(part(lagrangian,1),part(generators,kk),part(cogenerators,kk)) $
      write "fluxes  have been computed." $
    end
    else
      write "No variational problem." $
  end 
  else
      write "reliegen() has not been called yet!" $  
end $
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
procedure minors(m,k) $
begin
  integer k1,k2,k3,k4 $
  scalar nn,combrows,combcols,arow,acol,temp,allminors $   
  matrix tempmat(k,k) $
  nn:=length(m) $
  allminors:={} $
  if (k <= part(nn,1) and k <= part(nn,2)) then
  begin
    combrows:=combnorep(part(nn,1),k) $
    combcols:=combnorep(part(nn,2),k) $
    for k1:=1:length(combrows) do
    begin  
      arow:=part(combrows,k1) $
      for k2:=1:length(combcols) do
      begin
        acol:=part(combcols,k2) $
        for k3:=1:k do
          for k4:=1:k do
            tempmat(k3,k4):=m(part(arow,k3),part(acol,k4)) $
        temp:=det(tempmat) $
        if (temp neq 0) then allminors:=append(allminors,{temp}) $    
      end $
    end $   
  end
  else
    write "Order of requested minors too high." $
  clear tempmat $  
  return allminors $
end $
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
procedure inverselie(k) $
begin
  integer nr,nc $
  allminors:={} $
  if (steprelie geq 1) then
  begin
    if length(generators)>0 then
    begin  
      nr:=length(generators) $
      nc:=length(part(jet,1)) $
      if (k<=nr and k<=nc) then
      begin
        reliedistrib(generators) $     
        allminors:=minors(distribution,k) $
        write "The list 'allminors' has been computed." $
      end
      else
        write "Order of requested minors too high." $
    end 
    else
      write "Generators are not available." $
  end
  else  
    write "relieinit() has not been called yet!" $
  return allminors $
end $
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
relieheader() $
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%                              END OF RELIE.RED                                %%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
;end;
